// TFlat Menu 0.1b
// (C)Wikiz
//    wikiz@hdz.vejas.lt
//  slg HexaDecimalZ [ http://hdz.vejas.lt ]

// Short description

// This component isn't a real menu component. I is just used to
// assign painting events for a TmainMenu or TpopupMenu. Every Tmenu
// descendant (TmainMenu and TpopupMenu), added to a TFlatMenu.AssignedMenus
// collection, receive special events for measuring and painting. You cannot
// use CustomPaint events with a menu that has events received from TFlatMenu.

// To customize painting, use TFlatMenu's properies at design time. They have
// descriptive names. LinuxStyle makes a slight difference when painting TmainMenu:
// it defines how the top item of TmainMenu is painted - you'll notice what happens.

// Tested with Delphi5-6.

unit FlatMenu;


interface

uses
  Windows, Messages, SysUtils, Classes, Menus, graphics, extCtrls, forms;

const DEFAULT_RIGHT_MARGIN = 5;
      DEFAULT_LEFT_MARGIN = 16;
      DEFAULT_FOCUSED_FACE = $00E1EAEB;
      DEFAULT_LINE_COLOR = $008396A0;
      DEFAULT_NORMAL_FACE = clBtnFace;
      DEFAULT_DOWN_FACE = $00C5D6D9;
      DEFAULT_HIGHLIGHT_COLOR = clWhite;
      DEFAULT_SHADOW_COLOR = clBlack;
      DEFAULT_TEXT_VERT_MARGIN = 2;

type
  TFlatMenu = class;
  TMenuCollection =class;

  // holds one Tmenu component and manages Assigning/Unassigning
  // paint events
  TMenuCollectionItem = class(TCollectionItem)
   private
    FMenu: Tmenu;
    procedure SetMenu(const Value: Tmenu);
   protected
    function GetDisplayName: String; override;
    procedure AssignMenu ( const Value: Tmenu; Assign: boolean);
    procedure AssignTo(Dest: TPersistent); override;

  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    function FlatMenuComp: TflatMenu;
   published
    property Menu: Tmenu read FMenu write SetMenu;
  end;

  TMenuCollection = class(TCollection)
  private
    FMenu: TflatMenu;
    function GetItems(i: integer): TMenuCollectionItem;
  protected
    function GetOwner: TPersistent; override;
   public
    constructor Create(Acmp: TFlatMenu);
    property Menu: TflatMenu read FMenu;
    procedure DeleteLink ( const Acomp: Tcomponent );
    property Items[i: integer]: TMenuCollectionItem read GetItems;

  end;


  TFlatMenu = class(Tcomponent)
  private
    FFontActive: Tfont;
    FFontNormal: Tfont;
    FRightMargin: integer;
    FLeftMargin: integer;
    FDownFace: Tcolor;
    FnormalFace: Tcolor;
    FlineColor: Tcolor;
    FfocusedFace: Tcolor;
    FtextVert: integer;
    FAssignedMenus: TMenuCollection;
    FLinuxStyle: boolean;
    FShadowColor: Tcolor;
    FHighlightColor: Tcolor;
    FActive: boolean;
    function ItemMargin (item: TmenuItem; left: boolean): integer;
    procedure SetFontActive(const Value: Tfont);
    procedure SetFontNormal(const Value: Tfont);
    procedure SetAssignedMenus(const Value: TMenuCollection);
    function IsTopItem(item: TMenuItem): boolean;
    procedure SetActive(const Value: boolean);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation);
      override;
  public
    // painting events
    procedure DrawItem(Sender: TObject; ACanvas: TCanvas; ARect: TRect; State: TOwnerDrawState);
    procedure MeasureItem(Sender: TObject; ACanvas: TCanvas; var Width, Height: Integer);

    // recurse proc for assigning: use for new items, added at runtime
    procedure AssignMethods ( Aitem: TmenuItem; Assign: boolean );
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddMenu (const Amenu: Tmenu);


  published
    // the result is allways nil, but setting property assingns paint events for a menu
    property AssignedMenus: TMenuCollection read FAssignedMenus write SetAssignedMenus;
    property FontActive: Tfont read FFontActive write SetFontActive;
    property FontNormal: Tfont read FFontNormal write SetFontNormal;
    property LeftMargin: integer read FLeftMargin write FLeftMargin default DEFAULT_LEFT_MARGIN;
    property RightMargin: integer read FRightMargin write FRightMargin default DEFAULT_RIGHT_MARGIN;
    property TextVertMargin: integer read FtextVert write FTextVert default DEFAULT_TEXT_VERT_MARGIN;
    property FocusedFace: Tcolor read FfocusedFace write FfocusedFace default DEFAULT_FOCUSED_FACE;
    property LineColor: Tcolor read FlineColor write FlineColor default DEFAULT_LINE_COLOR;
    property NormalFace: Tcolor read FnormalFace write FnormalFace default DEFAULT_NORMAL_FACE;
    property DownFace: Tcolor read FDownFace write FDownFace default DEFAULT_DOWN_FACE;
    property LinuxStyle: boolean read FLinuxStyle write FLinuxStyle;
    property HighlightColor: Tcolor read FHighlightColor write FHighlightColor default DEFAULT_HIGHLIGHT_COLOR;
    property ShadowColor: Tcolor read FShadowColor write FShadowColor default DEFAULT_SHADOW_COLOR;
    property Active: boolean read FActive write SetActive;
  end;

procedure Register;

implementation
 uses math;

procedure Register;
begin
  RegisterComponents('FlatStyle', [TflatMenu]);
end;

{ TFlatMenu }


constructor TFlatMenu.Create(AOwner: TComponent);
begin
  inherited;
  FActive := true; 
  FAssignedMenus := TMenuCollection.Create(Self);

  FFontActive := TFont.Create;
  FFontActive.Style := [fsBold];
  FFontActive.name := 'Arial';
  FFontNormal := TFont.Create;
  FFontNormal.name := 'Arial';

  FLeftMargin := DEFAULT_LEFT_MARGIN;
  FRightMargin := DEFAULT_RIGHT_MARGIN;
  FDownFace := DEFAULT_DOWN_FACE;
  FnormalFace := DEFAULT_NORMAL_FACE;
  FlineColor := DEFAULT_LINE_COLOR;
  FfocusedFace := DEFAULT_FOCUSED_FACE;
  FtextVert := DEFAULT_TEXT_VERT_MARGIN;
  FHighlightColor := DEFAULT_HIGHLIGHT_COLOR;
  FShadowColor := DEFAULT_SHADOW_COLOR;
end;

destructor TFlatMenu.Destroy;
begin
  FAssignedMenus.Free;
  FFontActive.Free;
  FFontNormal.Free;
  inherited;
end;

function TFlatMenu.ItemMargin (item: TmenuItem; left: boolean): integer;
begin
 Result := FRightMargin;
 if not IsTopItem(item) and left
  then Result := FLeftMargin;
end;


function TFlatMenu.IsTopItem(item: TMenuItem): boolean;
begin
 Result := (item.GetParentMenu is TMainMenu) and (item.Parent.Parent=nil);
end;


procedure TFlatMenu.DrawItem(Sender: TObject; ACanvas: TCanvas; ARect: TRect; State: TOwnerDrawState);

 var it: TmenuItem;
     lLeft: integer;
     lLineTop: integer;
     lDown: boolean;
     lActive: boolean;
     lTopCol, lBottomCol, lColor: Tcolor;
     lTopItem: boolean;

     procedure PaintText ( Ax, Ay: integer; Atext: string; ArightsideX: boolean );
       var lPaintRect: trect;
     begin

       if (odDisabled in State)
        then Atext := StripHotkey(Atext);

       ACanvas.Brush.Style := bsSolid;
       if (odSelected in State) or (odHotLight in State)
        then ACanvas.Font.Assign (FFontActive)
        else ACanvas.Font.Assign(FFontNormal);
       if (odDefault in State) then ACanvas.Font.Style := ACanvas.Font.Style+[fsBold];


       if ArightsideX then Dec(Ax, ACanvas.TextWidth(Atext));
       if not lTopItem then inc(Ay, FtextVert-1);
       if lActive and not lTopItem then dec(Ay); // up one pixel

       lPaintRect := ARect;
       lPaintRect.Top := Ay;
       lPaintRect.Left := Ax;

        // just in case menu color and font color would be the same or similar
       if lTopItem and not lDown and not lActive
         then ACanvas.Font.Color := clMenuText;

       if odDisabled in State then begin
         ACanvas.Brush.Style := bsClear;
         ACanvas.Font.Color := clBtnHighlight;
         ACanvas.TextOut(Ax+1, Ay+1, Atext );
         ACanvas.Font.Color := clBtnShadow;
         ACanvas.TextOut(Ax, Ay, Atext );
       end
        else // normal
         DrawText(ACanvas.Handle, PChar(AText), Length(AText), lPaintRect,
                 DT_LEFT or DT_TOP or DT_NOCLIP);
     end;



begin
 it := TMenuItem(sender);
 lTopItem := IsTopItem(it); // main menu's top-level item, allways visible

 if it.IsLine then begin
    ACanvas.Pen.Width := 1;
    lLineTop := ARect.Top+(ARect.Bottom-ARect.Top) div 2;

    ACanvas.Brush.Color := FnormalFace;
    ACanvas.FillRect(ARect);
    ACanvas.Pen.Color := FlineColor;
    ACanvas.MoveTo(ARect.Left, lLineTop-1);
    ACanvas.LineTo(ARect.Right, lLineTop-1);
    ACanvas.Pen.Color := FlineColor;
    ACanvas.MoveTo(ARect.Left, lLineTop+1);
    ACanvas.LineTo(ARect.Right, lLineTop+1);
  end else begin
    lDown := not FLinuxStyle and lTopItem and (odSelected in state);
    lActive := not lDown and ((odSelected in state) or (odHotLight in state));

    lTopCol := FnormalFace;
    lBottomCol := FnormalFace;
    lColor := FnormalFace;

    if lDown then begin
     lTopCol := FShadowColor;
     lBottomCol := FHighlightColor;
     lColor := FDownFace;
    end;

    if lActive then begin
     lTopCol := FHighlightColor;
     lBottomCol := FShadowColor;
     lColor := FfocusedFace;
    end;

     // normal top item
    if not lActive and not lDown and lTopItem then begin
       lTopCol := clMenu;
       lBottomCol := clMenu;
       lColor := clMenu;
    end;



     Frame3D(ACanvas, ARect, lTopCol, lBottomCol, 1);
     ACanvas.Brush.Color := lColor;
     ACanvas.FillRect(ARect);


     // text
     lLeft := ARect.Left+ItemMargin(it, true);
     PaintText(lLeft, ARect.Top, it.Caption, false);
     lLeft := ARect.Right-ItemMargin(it, false);
     PaintText(lLeft, ARect.Top, ShortCutToText(it.ShortCut), true);


 end; // not is line
end;

procedure TFlatMenu.MeasureItem(Sender: TObject; ACanvas: TCanvas; var Width, Height: Integer);
  var it: TmenuItem;
      lTxt: string;

      procedure MeasureFor (const Afont: Tfont; const Atext: string);
       var w: integer;
      begin
        ACanvas.Font.Assign(Afont);
        w := ACanvas.TextWidth(Atext);
        if not IsTopItem(it) then Inc(w, ItemMargin(it, true)+ItemMargin(it, false));
        Width := Max(width, w);
        Height := Max(height, ACanvas.TextHeight(Atext)+FtextVert*2);
      end;

begin
 it := TMenuItem(sender);
 if it.IsLine then begin
   Height := 5; // this also can be done as option
 end else begin
   if it.Default then ACanvas.Font.Style := [fsBold];
   lTxt := StripHotkey(it.Caption);
   if it.ShortCut<>0 then lTxt := lTxt+' '+ShortCutToText(it.ShortCut);

   width := 0;
   height := 0;
   MeasureFor(FFontActive, lTxt);
   MeasureFor(FFontNormal, lTxt);
 end;  
end;


  // assigns only unasigned methods or unasigns only methods of self
procedure TFlatMenu.AssignMethods(Aitem: TmenuItem; Assign: boolean);
 var selfDraw, itemDraw: TAdvancedMenuDrawItemEvent;
     selfMeasure, itemMeasure: TMenuMeasureItemEvent;
     i: integer;
begin

 if Assign then begin
   if not Assigned(Aitem.OnAdvancedDrawItem) then Aitem.OnAdvancedDrawItem := DrawItem;
   if not Assigned(Aitem.OnMeasureItem) then Aitem.OnMeasureItem := MeasureItem;
 end else begin
   itemDraw := Aitem.OnAdvancedDrawItem;
   itemMeasure := Aitem.OnMeasureItem;
   selfDraw := DrawItem;
   selfMeasure := MeasureItem;
   if @itemDraw=@selfDraw then Aitem.OnAdvancedDrawItem := nil;
   if @itemMeasure=@selfMeasure then Aitem.OnAdvancedDrawItem := nil;
 end;

 // children
 for i := 0 to Aitem.Count-1 do AssignMethods(Aitem.Items[i], Assign);
end;



procedure TFlatMenu.SetFontActive(const Value: Tfont);
begin
  FFontActive.Assign(Value);
end;

procedure TFlatMenu.SetFontNormal(const Value: Tfont);
begin
  FFontNormal.Assign(Value);
end;



{ TMenuCollectionItem }

 // (un)registers a menu
procedure TMenuCollectionItem.AssignMenu(const Value: Tmenu; Assign: boolean);
begin
 if Value<>nil then
  TMenuCollection(Collection).Menu.AssignMethods(value.Items, Assign);

 if Assign then Value.FreeNotification(FlatMenuComp);
end;

constructor TMenuCollectionItem.Create(Collection: TCollection);
begin
  inherited;
  FMenu := nil;
end;

destructor TMenuCollectionItem.Destroy;
begin
  AssignMenu(FMenu, false);
  inherited;
end;

function TMenuCollectionItem.GetDisplayName: String;
begin
 if FMenu=nil
  then Result := 'Add a menu'
  else Result := FMenu.Name;
end;


procedure TMenuCollectionItem.SetMenu(const Value: Tmenu);
begin
  AssignMenu(FMenu, false);
  FMenu := Value;
  AssignMenu(FMenu, true);
  Fmenu.OwnerDraw := FlatMenuComp.Active;
end;

function TMenuCollectionItem.FlatMenuComp: TflatMenu;
begin
 Result := TMenuCollection(Collection).Menu;
end;


procedure TFlatMenu.SetAssignedMenus(const Value: TMenuCollection);
begin
  FAssignedMenus.Assign(Value);
end;

procedure TFlatMenu.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation=opRemove) and (Acomponent is Tmenu)
    then FAssignedMenus.DeleteLink(AComponent);
end;




procedure TFlatMenu.AddMenu(const Amenu: Tmenu);
 var it: TMenuCollectionItem;
begin
 it := TMenuCollectionItem(FAssignedMenus.Add);
 it.Menu := Amenu;
end;

procedure TFlatMenu.SetActive(const Value: boolean);
 var i: integer;
begin
  FActive := Value;
  for i := 0 to FAssignedMenus.Count-1 do
   if FAssignedMenus.Items[i].Menu<>nil
    then FAssignedMenus.Items[i].Menu.OwnerDraw := Value;
end;

{ TMenuCollection }

constructor TMenuCollection.Create(Acmp: Tflatmenu);
begin
 inherited Create(TMenuCollectionItem);
 FMenu := Acmp;
end;

procedure TMenuCollection.DeleteLink(const Acomp: Tcomponent);
 var i: integer;
begin
 i := 0;
 while i<Count do begin
   if Acomp= TMenuCollectionItem(Items[i]).Menu
    then Delete(i)
    else inc(i);
 end;
end;

function TMenuCollection.GetItems(i: integer): TMenuCollectionItem;
begin
 Result := TMenuCollectionItem(inherited items[i]);
end;

function TMenuCollection.GetOwner: TPersistent;
begin
 Result := Self.Menu;
end;

procedure TMenuCollectionItem.AssignTo(Dest: TPersistent);
begin
  if Dest is TMenuCollectionItem
   then TMenuCollectionItem(Dest).Menu := Menu
   else inherited;
end;


end.
